# -*- coding: utf-8 -*-
#
#   UDP: User Datagram Protocol
#
#                                Written in 2020 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2020 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import socket
from abc import ABC
from typing import Optional, Dict, Set

from startrek import Channel
from startrek import Connection, ConnectionDelegate
from startrek import BaseConnection
from startrek import BaseHub

from .channel import PackageChannel


class PackageHub(BaseHub, ABC):
    """ Base Package Hub """

    def __init__(self, delegate: ConnectionDelegate):
        super().__init__(delegate=delegate)
        # local => channel
        self.__channels: Dict[tuple, Channel] = {}

    def bind(self, address: tuple = None, host: str = None, port: int = 0):
        if address is None:
            address = (host, port)
        channel = self.__channels.get(address)
        if channel is None:
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
            sock.setblocking(True)
            sock.bind(address)
            sock.setblocking(False)
            channel = PackageChannel(sock=sock, remote=None, local=address)
            self.__channels[address] = channel

    def put_channel(self, channel: Channel):
        local = channel.local_address
        self.__channels[local] = channel

    # Override
    def _all_channels(self) -> Set[Channel]:
        return set(self.__channels.values())

    # Override
    def open_channel(self, remote: Optional[tuple], local: Optional[tuple]) -> Optional[Channel]:
        if local is None:
            for channel in self.__channels.values():
                return channel
            # channel not found
        else:
            return self.__channels.get(local)

    # Override
    def close_channel(self, channel: Channel):
        if channel is None or not channel.connected:
            # DON'T close bound socket channel
            return False
        else:
            self.__remove_channel(channel=channel)
        try:
            if channel.opened:
                channel.close()
            return True
        except socket.error:
            return False

    def __remove_channel(self, channel: Channel):
        local = channel.local_address
        if self.__channels.pop(local, None) == channel:
            # removed by key
            return True
        # remove by value
        for key in self.__channels:
            if self.__channels.get(key) == channel:
                self.__channels.pop(key, None)
                return True


class ServerHub(PackageHub):
    """ Package Server Hub """

    # Override
    def _create_connection(self, sock: Channel, remote: tuple, local: Optional[tuple]) -> Optional[Connection]:
        gate = self.delegate
        conn = BaseConnection(remote=remote, local=None, channel=sock, activated=False, delegate=gate, hub=self)
        conn.start()  # start FSM
        return conn


class ClientHub(PackageHub):
    """ Package Client Hub """

    # Override
    def _create_connection(self, sock: Channel, remote: tuple, local: Optional[tuple]) -> Optional[Connection]:
        gate = self.delegate
        conn = BaseConnection(remote=remote, local=None, channel=sock, activated=True, delegate=gate, hub=self)
        conn.start()  # start FSM
        return conn



# -*- coding: utf-8 -*-
#
#   MTP: Message Transfer Protocol
#
#                                Written in 2020 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2020 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Protocol
    ~~~~~~~~

    Data format in UDP payload
"""

from typing import Optional, Union

from ..ba import ByteArray, Data, Convert, MutableData

from .protocol import DataType, TransactionID

"""
    Package Header:

         0                   1                   2                   3
         0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |      'D'      |      'I'      |      'M'      | H-Len |  Type |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |                    Transaction ID (64 bits)                    
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
                             Transaction ID (64 bits)                   |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |               Fragment Pages (32 bits) OPTIONAL               |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |               Fragment Index (32 bits) OPTIONAL               |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |                 Body Length (32 bits) OPTIONAL                |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

        ** Magic Code **:
            Always be 'DIM'

        ** Header Length **:
            4 bits header length is the length of the header in 32 bit words.
            Note that the minimum value for a correct header is 3 (12 bytes).

        ** Data Type **:
            Indicates what kind of the body data is.

        ** Transaction ID **
            64 bits transaction ID is a random number for distinguishing
            different messages.

        ** Pages & Index **:
            If data type is a message fragment (or its respond),
            there is a field 'pages' following the transaction ID,
            which indicates the message was split to how many fragments;
            and there is another field 'index' following the 'pages'.

        ** Body Length **:
            Defined only for TCP stream.
            If transfer by UDP, no need to define the body's length.
"""


class Header(Data):
    """
        Max Length for message package body
        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        Each message package before split should not more than 1GB,
        so the max pages should not more than about 2,000,000.
    """
    MAX_BODY_LENGTH = 1024 * 1024 * 1024  # 1GB
    MAX_PAGES = 1024 * 1024 * 2           # 1GB

    MAGIC_CODE = b'DIM'

    def __init__(self, data: Union[bytes, bytearray, ByteArray],
                 data_type: DataType, sn: TransactionID, pages: int = 1, index: int = 0, body_length: int = -1):
        """
        Create package header

        :param data:      bytes, bytearray, Data or Header object
        :param data_type: package body data type
        :param sn:        transaction ID
        :param pages:     fragment count [OPTIONAL], default is 1
        :param index:     fragment index [OPTIONAL], default is 0
        :param body_length: length of body, default is -1 (unlimited)
        """
        if isinstance(data, ByteArray):
            super().__init__(buffer=data.buffer, offset=data.offset, size=data.size)
        else:
            super().__init__(buffer=data)
        self.__type = data_type
        self.__sn = sn
        self.__pages = pages
        self.__index = index
        self.__body_size = body_length

    def __str__(self) -> str:
        clazz = self.__class__.__name__
        dt = self.data_type
        bl = self.body_length
        return '<%s: %d, "%s" pages=%d, index=%d, body_len=%d />' % (clazz, self.size, dt, self.pages, self.index, bl)

    def __repr__(self) -> str:
        clazz = self.__class__.__name__
        dt = self.data_type
        bl = self.body_length
        return '<%s: %d, "%s" pages=%d, index=%d, body_len=%d />' % (clazz, self.size, dt, self.pages, self.index, bl)

    @property
    def data_type(self) -> DataType:
        return self.__type

    @property
    def is_response(self) -> bool:
        return self.__type.is_response

    @property
    def is_fragment(self) -> bool:
        return self.__type.is_fragment

    @property
    def is_command(self) -> bool:
        return self.__type.is_command

    @property
    def is_command_response(self) -> bool:
        return self.__type.is_command_response

    @property
    def is_message(self) -> bool:
        return self.__type.is_message

    @property
    def is_message_response(self) -> bool:
        return self.__type.is_message_response

    @property
    def is_message_fragment(self) -> bool:
        return self.__type.is_message_fragment

    @property
    def sn(self) -> TransactionID:
        return self.__sn

    @property
    def pages(self) -> int:
        """ fragment count """
        return self.__pages

    @property
    def index(self) -> int:
        """ fragment index"""
        return self.__index

    @property
    def body_length(self) -> int:
        return self.__body_size

    @classmethod
    def parse(cls, data: ByteArray):  # -> Header:
        data_type = get_data_type(data=data)
        if data_type is None:
            # not a DIM package?
            return None
        head_len = get_header_length(data=data)
        if data.size < head_len:
            # waiting for more data
            return None
        pages = 1
        index = 0
        body_len = -1
        if head_len == 4:
            """ simple header (for UDP only)
                +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
                |      'D'      |      'I'      |      'M'      | H-Len |  Type |
                +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            """
            sn = TransactionID.ZERO
        elif head_len == 8:
            """ simple header with body length
                +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
                |      'D'      |      'I'      |      'M'      | H-Len |  Type |
                +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
                |                 Body Length (32 bits) OPTIONAL                |
                +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            """
            sn = TransactionID.ZERO
            body_len = Convert.int32_from_data(data=data, start=4)
        elif head_len == 12:
            """ command/message header without body length (for UDP only)
                +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
                |      'D'      |      'I'      |      'M'      | H-Len |  Type |
                +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
                |                    Transaction ID (64 bits)                    
                +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
                                     Transaction ID (64 bits)                   |
                +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            """
            sn = TransactionID.from_data(data=data.slice(start=4, end=12))
        elif head_len == 16:
            """ command/message header with body length
                +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
                |      'D'      |      'I'      |      'M'      | H-Len |  Type |
                +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
                |                    Transaction ID (64 bits)                    
                +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
                                     Transaction ID (64 bits)                   |
                +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
                |                 Body Length (32 bits) OPTIONAL                |
                +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

            """
            sn = TransactionID.from_data(data=data.slice(start=4, end=12))
            body_len = Convert.int32_from_data(data=data, start=12)
        elif head_len == 20:
            """ fragment header without body length (for UDP only)
                +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
                |      'D'      |      'I'      |      'M'      | H-Len |  Type |
                +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
                |                    Transaction ID (64 bits)                    
                +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
                                     Transaction ID (64 bits)                   |
                +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
                |               Fragment Pages (32 bits) OPTIONAL               |
                +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
                |               Fragment Index (32 bits) OPTIONAL               |
                +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            """
            sn = TransactionID.from_data(data=data.slice(start=4, end=12))
            pages = Convert.int32_from_data(data=data, start=12)
            index = Convert.int32_from_data(data=data, start=16)
        elif head_len == 24:
            """ fragment header with body length
                +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
                |      'D'      |      'I'      |      'M'      | H-Len |  Type |
                +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
                |                    Transaction ID (64 bits)                    
                +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
                                     Transaction ID (64 bits)                   |
                +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
                |               Fragment Pages (32 bits) OPTIONAL               |
                +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
                |               Fragment Index (32 bits) OPTIONAL               |
                +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
                |                 Body Length (32 bits) OPTIONAL                |
                +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            """
            sn = TransactionID.from_data(data=data.slice(start=4, end=12))
            pages = Convert.int32_from_data(data=data, start=12)
            index = Convert.int32_from_data(data=data, start=16)
            body_len = Convert.int32_from_data(data=data, start=20)
        else:
            # raise ValueError('header length error: %d' % head_len)
            return None
        assert sn is not None, 'Transaction ID error: %s' % data
        if pages < 1 or pages > cls.MAX_PAGES:
            # raise ValueError('pages error: %d' % pages)
            return None
        if index < 0 or index >= pages:
            # raise ValueError('index error: %d' % index)
            return None
        if body_len < -1 or body_len > cls.MAX_BODY_LENGTH:
            # raise ValueError('body length error: %d' % body_len)
            return None
        # create header
        return cls(data=data.slice(start=0, end=head_len),
                   data_type=data_type, sn=sn, pages=pages, index=index, body_length=body_len)

    @classmethod
    def new(cls, data_type: DataType, sn: TransactionID = None, pages: int = 1, index: int = 0, body_length: int = -1):
        assert data_type is not None, 'data type should not be None'
        assert 0 <= index < pages, 'pages error: %d, index: %d' % (pages, index)
        assert -1 <= body_length < cls.MAX_BODY_LENGTH, 'body length error: %d' % body_length
        head_len = 4  # in bytes
        # 1. transaction ID
        if sn is None:
            # generate transaction ID
            sn = TransactionID.generate()
            head_len += sn.size  # 8 bytes
        elif sn == TransactionID.ZERO:
            # simple header
            sn = None
        else:
            head_len += sn.size  # 8 bytes
        options = None
        # 2. pages & index
        if pages > 1:
            # message fragment (or its respond)
            d1 = Convert.uint32data_from_value(value=pages)
            d2 = Convert.uint32data_from_value(value=index)
            options = d1.concat(d2)
            head_len += d1.size + d2.size  # 8 bytes
        # 3. body length
        if body_length >= 0:
            # for TCP
            d3 = Convert.uint32data_from_value(value=body_length)
            if options is None:
                options = d3
            else:
                options = options.concat(d3)
            head_len += d3.size  # 4 bytes
        # generate header data
        hl_ty = (head_len << 2) | (data_type.value & 0x0F)
        data = MutableData(capacity=head_len)
        data.append(source=cls.MAGIC_CODE)  # 'DIM'
        data.push(element=hl_ty)
        if sn is not None:
            data.append(source=sn)
        if options is not None:
            data.append(source=options)
        return cls(data=data, data_type=data_type, sn=sn, pages=pages, index=index, body_length=body_length)


def get_data_type(data: ByteArray) -> Optional[DataType]:
    if data.size < 4:
        # waiting for more data
        return None
    buffer = data.buffer
    offset = data.offset
    if buffer[offset:(offset+3)] != Header.MAGIC_CODE:
        # raise ValueError('not a DIM package: %s' % data)
        return None
    return DataType.from_data(data=data.slice(start=3, end=4))


def get_header_length(data: ByteArray) -> int:
    # header length shared the byte with data type
    ch = data.get_byte(index=3)
    hl = (ch & 0xF0) >> 4
    if hl < 1 or hl > 6:
        return 0
    else:
        return hl << 2  # in bytes



# -*- coding: utf-8 -*-
#
#   Star Trek: Interstellar Transport
#
#                                Written in 2021 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2021 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from typing import List, Optional

from startrek import Arrival, ArrivalShip
from startrek import Departure, DepartureShip, DeparturePriority
from startrek import ShipDelegate
from startrek import StarDocker, StarGate

from .ba import Data
from .mtp import DataType, TransactionID, Package, Packer


class PackageArrival(ArrivalShip):

    def __init__(self, pack: Package):
        super().__init__()
        head = pack.head
        self.__sn = head.sn
        if head.is_fragment:
            self.__packer = Packer(sn=head.sn, pages=head.pages)
            self.__completed = self.__packer.insert(fragment=pack)
        else:
            self.__packer = None
            self.__completed = pack

    @property
    def package(self) -> Optional[Package]:
        return self.__completed

    @property
    def fragments(self) -> Optional[List[Package]]:
        packer = self.__packer
        if packer is not None:
            assert isinstance(packer, Packer), 'packer error: %s' % packer
            return packer.fragments

    @property  # Override
    def sn(self) -> TransactionID:
        return self.__sn

    # Override
    def assemble(self, ship):  # -> Optional[PackageArrival]:
        if self.__completed is None and ship is not self:
            packer = self.__packer
            # assert isinstance(packer, Packer), 'packer error: %s' % packer
            # assert isinstance(ship, PackageArrival), 'arrival ship error: %s' % ship
            fragments = ship.fragments
            # assert fragments is not None and len(fragments) > 0, 'fragments error: %s' % ship
            for item in fragments:
                self.__completed = packer.insert(fragment=item)
        if self.__completed is not None:
            return self


class PackageDeparture(DepartureShip):

    def __init__(self, pack: Package, priority: int = 0, delegate: ShipDelegate = None):
        super().__init__(priority=priority, delegate=delegate)
        self.__completed = pack
        self.__packages = self._split_package(pack=pack)
        self.__fragments: List[bytes] = []

    # noinspection PyMethodMayBeStatic
    def _split_package(self, pack: Package) -> List[Package]:
        if pack.is_message:
            return Packer.split(package=pack)
        else:
            return [pack]

    @property
    def package(self) -> Package:
        return self.__completed

    @property  # Override
    def sn(self) -> TransactionID:
        return self.__completed.head.sn

    @property  # Override
    def fragments(self) -> List[bytes]:
        if len(self.__fragments) == 0 and len(self.__packages) > 0:
            for item in self.__packages:
                self.__fragments.append(item.get_bytes())
        return self.__fragments

    # Override
    def check_response(self, ship: Arrival) -> bool:
        count = 0
        assert isinstance(ship, PackageArrival), 'arrival ship error: %s' % ship
        fragments = ship.fragments
        if fragments is None:
            # it's a completed data package
            pack = ship.package
            if self.__remove_page(index=pack.head.index):
                count += 1
        else:
            for pack in fragments:
                if self.__remove_page(index=pack.head.index):
                    count += 1
        if count > 0:
            self.__fragments.clear()
            return len(self.__packages) == 0

    def __remove_page(self, index: int) -> bool:
        for pack in self.__packages:
            if pack.head.index == index:
                # got it
                self.__packages.remove(pack)
                return True


class PackageDocker(StarDocker):

    def __init__(self, remote: tuple, local: Optional[tuple], gate: StarGate):
        super().__init__(remote=remote, local=local, gate=gate)

    # noinspection PyMethodMayBeStatic
    def _parse_package(self, data: Optional[bytes]) -> Optional[Package]:
        if data is not None:  # and len(data) > 0:
            return Package.parse(data=Data(buffer=data))

    # noinspection PyMethodMayBeStatic
    def _create_arrival(self, pack: Package) -> Arrival:
        return PackageArrival(pack=pack)

    # noinspection PyMethodMayBeStatic
    def _create_departure(self, pack: Package, priority: int = 0, delegate: ShipDelegate = None) -> Departure:
        return PackageDeparture(pack=pack, priority=priority, delegate=delegate)

    # Override
    def _next_departure(self, now: int) -> Optional[Departure]:
        outgo = super()._next_departure(now=now)
        if outgo is not None:
            self._retry_departure(ship=outgo)
        return outgo

    def _retry_departure(self, ship: Departure):
        if ship.retries >= DepartureShip.MAX_RETRIES:
            # last try
            return False
        if isinstance(ship, PackageDeparture):
            pack = ship.package
            if not pack.is_response:
                # put back for next retry
                return self.append_departure(ship=ship)

    # Override
    def _get_arrival(self, data: bytes) -> Optional[Arrival]:
        pack = self._parse_package(data=data)
        if pack is not None:  # and pack.body.size > 0:
            return self._create_arrival(pack=pack)

    # Override
    def _check_arrival(self, ship: Arrival) -> Optional[Arrival]:
        assert isinstance(ship, PackageArrival), 'arrival ship error: %s' % ship
        pack = ship.package
        if pack is None:
            fragments = ship.fragments
            if fragments is None or len(fragments) == 0:
                raise ValueError('fragments error: %s' % ship)
            # each ship can carry one fragment only
            pack = fragments[0]
        # check data type in package header
        head = pack.head
        body = pack.body
        data_type = head.data_type
        if data_type.is_command_response:
            # process CommandResponse
            #       'PONG'
            #       'OK'
            self._check_response(ship=ship)
            if body == PONG or body == OK:
                # command responded
                return None
            # extra data in CommandResponse?
            # let the caller to process it
        elif data_type.is_command:
            # process Command:
            #       'PING'
            #       '...'
            if body == PING:
                # PING -> PONG
                self._respond_command(sn=head.sn, body=PONG)
                return None
            else:
                # respond for Command
                self._respond_command(sn=head.sn, body=OK)
            # Unknown Command?
            # let the caller to process it
        elif data_type.is_message_response:
            # process MessageResponse
            #       'OK'
            #       'AGAIN'
            if body == AGAIN:
                # TODO: reset retries?
                return None
            self._check_response(ship=ship)
            if body == OK:
                # message responded
                return None
            # extra data in MessageResponse?
            # let the caller to process it
        else:
            # respond for Message/Fragment
            self._respond_message(sn=head.sn, pages=head.pages, index=head.index)
            if data_type.is_message_fragment:
                # assemble MessageFragment with cached fragments to completed Message
                # let the caller to process the completed message
                return self._assemble_arrival(ship=ship)
            assert data_type.is_message, 'unknown data type: %s' % data_type
            # let the caller to process the message

        if body.size == 4:
            if body == NOOP:
                # do nothing
                return None
            elif body == PING or body == PONG:
                # FIXME: these bodies should be in a Command
                # ignore them
                return None
        return ship

    # protected
    def _respond_command(self, sn: TransactionID, body: bytes):
        pack = Package.new(data_type=DataType.COMMAND_RESPONSE, sn=sn, body=Data(buffer=body))
        self.send_package(pack=pack)

    # protected
    def _respond_message(self, sn: TransactionID, pages: int, index: int):
        pack = Package.new(data_type=DataType.MESSAGE_RESPONSE, sn=sn, pages=pages, index=index, body=Data(buffer=OK))
        self.send_package(pack=pack)

    def send_package(self, pack: Package, priority: int = 0, delegate: ShipDelegate = None):
        if delegate is None:
            delegate = self.delegate
        ship = self._create_departure(pack=pack, priority=priority, delegate=delegate)
        self.append_departure(ship=ship)

    def send_ship(self, ship: Departure):
        self.append_departure(ship=ship)

    # Override
    def pack(self, payload: bytes, priority: int = 0, delegate: ShipDelegate = None) -> Departure:
        pkg = Package.new(data_type=DataType.MESSAGE, body=Data(buffer=payload))
        return self._create_departure(pack=pkg, priority=priority, delegate=delegate)

    # Override
    def heartbeat(self):
        pkg = Package.new(data_type=DataType.COMMAND, body=Data(buffer=PING))
        outgo = self._create_departure(pack=pkg, priority=DeparturePriority.SLOWER)
        self.append_departure(ship=outgo)


PING = b'PING'
PONG = b'PONG'
NOOP = b'NOOP'
OK = b'OK'
AGAIN = b'AGAIN'



# -*- coding: utf-8 -*-
#
#   UDP: User Datagram Protocol
#
#                                Written in 2020 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2020 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import socket

from startrek.net.base_channel import sendall
from startrek import BaseChannel


class PackageChannel(BaseChannel):
    """ Discrete Package Channel """

    # Override
    def receive(self, max_len: int) -> (bytes, tuple):
        sock = self.sock
        if sock is None:
            raise socket.error('socket lost, cannot read data')
        remote = self.remote_address
        # try to receive data
        try:
            if remote is None:
                # UDP receiving
                data, remote = sock.recvfrom(max_len)
            else:
                # connected (TCP/UDP)
                data = sock.recv(max_len)
        except socket.error as error:
            # the socket will raise 'Resource temporarily unavailable'
            # when received nothing in non-blocking mode,
            # here we should ignore this exception.
            if not self.blocking:
                if error.strerror == 'Resource temporarily unavailable':
                    # received nothing
                    return None, None
            raise error
        # check data
        if data is None or len(data) == 0:
            # in blocking mode, the socket will wait until received something,
            # but if timeout was set, it will return None too, it's normal;
            # otherwise, we know the connection was lost.
            if sock.gettimeout() is None:  # and self.blocking:
                raise socket.error('remote peer reset socket')
        return data, remote

    # Override
    def send(self, data: bytes, target: tuple) -> int:
        sock = self.sock
        if sock is None:
            raise socket.error('socket lost, cannot send data: %d byte(s)' % len(data))
        remote = self.remote_address
        if remote is None:
            # UDP sending
            return sock.sendto(data, target)
        else:
            # connected (TCP/UDP)
            assert target is None or target == remote,\
                'the target must equal to remote address: %s, %s' % (target, remote)
            # return sock.sendall(data)
            return sendall(data=data, sock=sock)



# -*- coding: utf-8 -*-
#
#   BA: Byte Array
#
#                                Written in 2020 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2020 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Byte Array
    ~~~~~~~~~~
"""

from .array import ByteArray, MutableByteArray
from .array import Endian, IntegerData
from .helper import *

from .data import Data
from .mutable import MutableData
from .integer import IntData, UInt8Data, UInt16Data, UInt32Data, VarIntData

from .convert import Convert

name = "BA"

__author__ = 'Albert Moky'

__all__ = [
    'ByteArray', 'MutableByteArray',
    'Endian', 'IntegerData',

    'Data', 'MutableData',
    'IntData', 'UInt8Data', 'UInt16Data', 'UInt32Data', 'VarIntData',
    'Convert',
]



# -*- coding: utf-8 -*-
#
#   BA: Byte Array
#
#                                Written in 2020 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2020 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from typing import Optional, Union

from .array import ByteArray, IntegerData, Endian
from .integer import UInt16Data, UInt32Data


class Convert:
    """
        Network Byte Order Converter
        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    """

    @classmethod
    def int_from_data(cls, data: Union[bytes, bytearray, ByteArray], start: int, size: int) -> int:
        return IntegerData.get_value(source=data, start=start, size=size, endian=Endian.BIG_ENDIAN)

    @classmethod
    def int16_from_data(cls, data: Union[bytes, bytearray, ByteArray], start: int = 0) -> int:
        return cls.int_from_data(data=data, start=start, size=2)

    @classmethod
    def int32_from_data(cls, data: Union[bytes, bytearray, ByteArray], start: int = 0) -> int:
        return cls.int_from_data(data=data, start=start, size=4)

    #
    #   UInt16Data
    #

    @classmethod
    def uint16data_from_value(cls, value: int) -> UInt16Data:
        return UInt16Data.from_int(value=value, endian=Endian.BIG_ENDIAN)

    @classmethod
    def uint16data_from_data(cls, data: Union[bytes, bytearray, ByteArray], start: int = 0) -> Optional[UInt16Data]:
        data = cut(data=data, start=start)
        return UInt16Data.from_data(data=data, endian=Endian.BIG_ENDIAN)

    #
    #   UInt32Data
    #

    @classmethod
    def uint32data_from_value(cls, value: int) -> UInt32Data:
        return UInt32Data.from_int(value=value, endian=Endian.BIG_ENDIAN)

    @classmethod
    def uint32data_from_data(cls, data: Union[bytes, bytearray, ByteArray], start: int = 0) -> Optional[UInt32Data]:
        data = cut(data=data, start=start)
        return UInt32Data.from_data(data=data, endian=Endian.BIG_ENDIAN)


def cut(data: Union[bytes, bytearray, ByteArray], start: int = 0) -> Union[bytes, bytearray, ByteArray]:
    if start == 0:
        return data
    elif isinstance(data, ByteArray):
        return data.slice(start=start)
    else:
        return data[start:]



# -*- coding: utf-8 -*-
#
#   BA: Byte Array
#
#                                Written in 2020 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2020 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import binascii
import random
from typing import Union

from .array import Endian


def hex_encode(data: bytes) -> str:
    return binascii.b2a_hex(data).decode('utf-8')


def hex_decode(string: str) -> bytes:
    return binascii.a2b_hex(string)


def random_bytes(size: int) -> bytes:
    a = [random.choice('0123456789ABCDEF') for _ in range(size << 1)]
    return hex_decode(''.join(a))


def array_copy(src: Union[bytes, bytearray], src_pos: int, dest: bytearray, dest_pos: int, length: int):
    """
    Copies an array from the specified source array, beginning at the
    specified position, to the specified position of the destination array.
    A sub-sequence of array components are copied from the source array
    referenced by src to the destination array referenced by dest.
    The number of components copied is equal to the length argument.
    The components at positions srcPos through srcPos+length-1 in the
    source array are copied into positions destPos through destPos+length-1,
    respectively, of the destination array.

    :param src:      the source array.
    :param src_pos:  starting position in the source array.
    :param dest:     the destination array.
    :param dest_pos: starting position in the destination data.
    :param length:   the number of array elements to be copied.
    :return:
    """
    if src is dest:
        # same buffer
        if src_pos == dest_pos:
            # same position, do nothing
            return False
        #
        #   now we have:
        #       start1 = src_pos,  end1 = src_pos + length,
        #       start2 = dest_pos, end2 = dest_pos + length,
        #       and length > 0
        #
        #   check for intersections:
        #       1. start1 < end1 <= start2 < end2
        #          src:  ********........
        #          dest: ........########
        #       2. start1 < start2 < end1 < end2   (intersected, needs temporary buffer)
        #          src:  ..********......
        #          dest: ......########..
        #       3. start1 == start2 < end1 == end2 (intersected, but skipped)
        #          src:  ....********....
        #          dest: ....########....
        #       4. start2 < start1 < end2 < end1   (intersected, but doesn't need temporary buffer)
        #          src:  ......********..
        #          dest: ..########......
        #       5. start2 < end2 <= start1 < end1
        #          src:  ........********
        #          dest: ########........
        #
        src_end = src_pos + length
        if src_pos < dest_pos < src_end:
            # copy to a temporary buffer
            src = src[src_pos:src_end]
            src_pos = 0
    for i in range(length):
        dest[dest_pos+i] = src[src_pos+i]
    return True


def array_equal(left_buffer: Union[bytes, bytearray], left_offset: int, left_size: int,
                right_buffer: Union[bytes, bytearray], right_offset: int, right_size: int) -> bool:
    """ Check whether buffers equal with range [offset, offset + size) """
    if left_size != right_size:
        return False
    if left_buffer is right_buffer:
        # same buffer
        if left_offset == right_offset and left_size == right_size:
            # same range
            return True
    pos1 = left_offset + left_size - 1
    pos2 = right_offset + right_size - 1
    while pos2 >= right_offset:
        if left_buffer[pos1] != right_buffer[pos2]:
            # not matched
            return False
        pos1 -= 1
        pos2 -= 1
    # all items matched
    return True


def array_hash(buffer: Union[bytes, bytearray], offset: int, size: int) -> int:
    """ Calculate hash code for buffer with range [offset, offset + size) """
    result = 1
    start = offset
    end = offset + size
    while start < end:
        result = (result << 5) - result + buffer[start]
        start += 1
    return result


def array_concat(left_buffer: Union[bytes, bytearray],
                 left_offset: int, left_size: int,
                 right_buffer: Union[bytes, bytearray],
                 right_offset: int, right_size: int) -> (Union[bytes, bytearray], int, int):
    """ Concat two data with range [start, end) and return (buffer, offset, size) """
    if left_size == 0:
        return right_buffer, right_offset, right_size
    elif right_size == 0:
        return left_buffer, left_offset, left_size
    left_end = left_offset + left_size
    right_end = right_offset + right_size
    if left_buffer is right_buffer:
        # same buffer
        if left_end == right_offset:
            # sticky data
            return left_buffer, left_offset, left_size + right_size
    # slice buffers
    if 0 < left_offset or left_end < len(left_buffer):
        left_buffer = left_buffer[left_offset:left_end]
    if 0 < right_offset or right_end < len(right_buffer):
        right_buffer = right_buffer[right_offset:right_end]
    # check types
    if isinstance(left_buffer, bytearray):
        if not isinstance(right_buffer, bytearray):
            # bytearray + bytes
            right_buffer = bytearray(right_buffer)
    elif isinstance(right_buffer, bytearray):
        # bytes + bytearray
        left_buffer = bytearray(left_buffer)
    return left_buffer + right_buffer, 0, left_size + right_size


def array_find(sub_buffer: Union[bytes, bytearray], sub_offset: int, sub_size: int,
               buffer: Union[bytes, bytearray], offset: int, size: int) -> int:
    """ Searching sub data in buffer with range [offset, offset + size) """
    if sub_size <= 0 or sub_size > size:
        # the sub data is empty or too large
        return -1
    # pre check
    start = offset
    end = offset + size
    found = -1
    if buffer is sub_buffer:
        # same buffer
        if offset == sub_offset:
            # the sub.offset matched the start position,
            # it's surely the first position the sub data appeared.
            return 0
        if offset < sub_offset <= (end - sub_size):
            # if sub.offset is in range (start, end - sub.size],
            # the position (sub.offset - self.offset) is matched,
            # but we cannot confirm this is the first position the sub data appeared,
            # so we still need to do searching in range [start, sub.offset + sub.size - 1).
            found = sub_offset - offset
            end = sub_offset + sub_size - 1
    # TODO: Boyer-Moore Searching
    sub_start = sub_offset
    sub_end = sub_offset + sub_size
    if 0 < sub_start or sub_end < len(sub_buffer):
        sub = sub_buffer[sub_start:sub_end]
    else:
        sub = sub_buffer
    # do searching
    pos = buffer.find(sub, start, end)
    if pos == -1:
        return found
    else:
        return pos - offset


def array_set(index: int, value: int, buffer: bytearray, offset: int, size: int) -> int:
    """ Set value at index, return new size """
    pos = offset + index
    tail = offset + size
    buf_len = len(buffer)
    if buf_len <= pos:
        # the target position is outside of the current buffer
        buffer.extend(bytearray(pos - buf_len + 1))
    # set value
    buffer[pos] = value & 0xFF
    # clean the gaps if exist?
    if tail < pos:
        for i in range(tail, pos):
            buffer[i] = 0
    # return the maximum size
    if index < size:
        return size
    else:
        return index + 1


def array_update(index: int, src: Union[bytes, bytearray], src_offset: int, src_size: int,
                 buffer: bytearray, offset: int, size: int) -> int:
    """ Update src from index, return new size """
    src_end = src_offset + src_size
    if src_offset > 0 or src_end < len(src):
        src = src[src_offset:src_end]
    start = offset + index
    end = start + src_size
    tail = offset + size
    buf_len = len(buffer)
    #
    #   now we have:
    #       start = offset + index
    #       end = offset + index + src_size
    #       tail = offset + size
    #       src_size > 0
    #
    #   check for intersections:
    #       1. buf_len <= start < end
    #          src:                  ****
    #          dest: ########........
    #       2. start < buf_len < end
    #          src:                ****
    #          dest: ########........
    #       3. start < end <= buf_len
    #          src:              ****
    #          dest: ########........
    #
    if buf_len <= start:
        # the target range is outside of the current buffer
        if buf_len < start:
            # append the gaps
            buffer.extend(bytearray(start - buf_len))
        # append source data to tail
        buffer.extend(src)
    elif buf_len < end:
        # the target range is partially outside of the current buffer
        mid = buf_len - start
        for i in range(mid):
            # copy the left part
            buffer[start+i] = src[i]
        # append the right part to tail
        buffer.extend(src[mid:])
    else:
        # the target range is totally inside the current buffer
        for i in range(src_size):
            buffer[start+i] = src[i]
    # clean the gaps if exist?
    if tail < start:
        for i in range(tail, start):
            buffer[i] = 0
    # return the new size
    if tail < end:
        return end - offset
    else:
        return tail - offset


def array_insert(index: int, src: Union[bytes, bytearray], src_offset: int, src_size: int,
                 buffer: bytearray, offset: int, size: int) -> int:
    """ Insert src at index, return new size """
    if index < size:
        start = offset + index
        for i in range(src_size):
            buffer.insert(start+i, src[src_offset+i])
        return size + src_size
    else:
        return array_update(index=index,
                            src=src, src_offset=src_offset, src_size=src_size,
                            buffer=buffer, offset=offset, size=size)


def array_remove(index: int, buffer: bytearray, offset: int, size: int) -> (int, int, int):
    """ Remove element at index, return its value and new offset & size """
    pos = offset + index
    if index == 0:
        # remove the first element
        return buffer[pos], offset + 1, size - 1
    elif index == (size - 1):
        # remove the last element
        return buffer[pos], offset, size - 1
    elif 0 < index < (size - 1):
        return buffer.pop(pos), offset, size - 1
    else:
        raise IndexError('error index: %d, size: %d' % (index, size))


#
#  Converting
#

def int_from_buffer(buffer: Union[bytes, bytearray], offset: int, size: int, endian: Endian) -> int:
    """
    Get integer value from data buffer with range [offset, offset + size)

    :param buffer: data buffer
    :param offset: data view offset
    :param size:   data view size
    :param endian  byte order
    :return: int value
    """
    value = 0
    if endian == Endian.LITTLE_ENDIAN:
        # [12 34 56 78] => 0x78563412
        pos = offset + size - 1
        while pos >= offset:
            value = (value << 8) | (buffer[pos] & 0xFF)
            pos -= 1
    elif endian == Endian.BIG_ENDIAN:
        # [12 34 56 78] => 0x12345678
        pos = offset
        end = offset + size
        while pos < end:
            value = (value << 8) | (buffer[pos] & 0xFF)
            pos += 1
    return value


def int_to_buffer(value: int, buffer: bytearray, offset: int, size: int, endian: Endian):
    """
    Set integer value into data buffer with size

    :param value:  int value
    :param buffer: data buffer
    :param offset: data view offset
    :param size:   data view size
    :param endian  byte order
    """
    if endian == Endian.LITTLE_ENDIAN:
        # 0x12345678 => [78 56 34 12]
        pos = offset
        end = offset + size
        while pos < end:
            buffer[pos] = value & 0xFF
            value >>= 8
            pos += 1
    elif endian == Endian.BIG_ENDIAN:
        # 0x12345678 => [12 34 56 78]
        pos = offset + size - 1
        while pos >= offset:
            buffer[pos] = value & 0xFF
            value >>= 8
            pos -= 1


"""
    VarInt
    ~~~~~~

    0zzzzzzz                   <-> 0zzzzzzz
    00yyyyyy yzzzzzzz          <-> 1zzzzzzz 0yyyyyyy
    000xxxxx xxyyyyyy yzzzzzzz <-> 1zzzzzzz 1yyyyyyy 0xxxxxxx
    ...
"""


def varint_from_buffer(buffer: Union[bytes, bytearray], offset: int, size: int) -> (int, int):
    """
    Get integer value from variable data buffer with range [offset, offset + size)

    :param buffer: data buffer
    :param offset: data view offset
    :param size:   data view size
    :return: int value, data length
    """
    value = 0
    bits = 0
    pos = offset
    end = offset + size
    ch = 0x80
    while (ch & 0x80) != 0:
        if pos >= end:
            # raise ValueError('out of range: [%d, %d)' % (offset, end))
            return 0, 0
        ch = buffer[pos]
        value |= (ch & 0x7F) << bits
        bits += 7
        pos += 1
    # return (value, length)
    return value, pos - offset


def varint_to_buffer(value: int, buffer: bytearray, offset: int, size: int) -> int:
    """
    Set integer value into variable data buffer with range [offset, offset + size)

    :param value:  int value
    :param buffer: data buffer
    :param offset: data view offset
    :param size:   data view size
    :return: data length
    """
    pos = offset
    end = offset + size
    while value > 0x7F:
        if pos >= end:
            raise OverflowError('out of range: [%d, %d), value=%d' % (offset, end, value))
        buffer[pos] = (value & 0x7F) | 0x80
        value >>= 7
        pos += 1
    if pos >= end:
        raise OverflowError('out of range: [%d, %d), value=%d' % (offset, end, value))
    buffer[pos] = (value & 0x7F)
    return pos - offset + 1



# -*- coding: utf-8 -*-
#
#   MTP: Message Transfer Protocol
#
#                                Written in 2020 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2020 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Protocol
    ~~~~~~~~

    Data format in UDP payload
"""

import threading
from typing import Union

from ..ba import ByteArray, Data, Convert, UInt8Data

"""
    Data Type:
    
          0   1   2   3   4   5   6   7
        +---+---+---+---+---+---+---+---+
        |               | F |   | M | A |
        |               | R |   | S | C |
        |               | G |   | G | K |
        +---+---+---+---+---+---+---+---+
    
        Command                  : 0x00 (0000 0000)
        Command Respond          : 0x01 (0000 0001)
        Message                  : 0x02 (0000 0010)
        Message Respond          : 0x03 (0000 0011)
        Message Fragment         : 0x0A (0000 1010)
"""


class DataType(UInt8Data):

    def __init__(self, data: Union[bytes, bytearray, ByteArray], value: int, name: str):
        super().__init__(data=data, value=value)
        self.__name = name

    def __str__(self) -> str:
        return self.__name

    def __repr__(self) -> str:
        return self.__name

    @property
    def name(self) -> str:
        return self.__name

    @property
    def is_response(self) -> bool:
        return (self.value & 0x01) != 0

    @property
    def is_fragment(self) -> bool:
        return (self.value & 0x08) != 0

    @property
    def is_command(self) -> bool:
        return (self.value & 0x0F) == 0x00

    @property
    def is_command_response(self) -> bool:
        return (self.value & 0x0F) == 0x01

    @property
    def is_message(self) -> bool:
        return (self.value & 0x0F) == 0x02

    @property
    def is_message_response(self) -> bool:
        return (self.value & 0x0F) == 0x03

    @property
    def is_message_fragment(self) -> bool:
        return (self.value & 0x0F) == 0x0A

    #
    #   Factories
    #

    @classmethod
    def from_data(cls, data: Union[bytes, bytearray, ByteArray]):
        if not isinstance(data, UInt8Data):
            data = UInt8Data.from_data(data=data)
        if data is not None:
            return cls.__get(data=data)

    @classmethod
    def from_int(cls, value: int):
        data = UInt8Data.from_int(value=value)
        return cls.__get(data=data)

    @classmethod
    def __get(cls, data: UInt8Data):
        value = data.value & 0x0F
        fixed = cls.__data_types.get(value)
        if fixed is not None:
            # assert isinstance(fixed, DataType)
            return cls(data=data, value=value, name=fixed.name)

    @classmethod
    def cache(cls, value: int, data_type):
        cls.__data_types[value] = data_type

    __data_types = {}  # int -> DataType

    # fixed types
    COMMAND = None
    COMMAND_RESPONSE = None
    MESSAGE = None
    MESSAGE_RESPONSE = None
    MESSAGE_FRAGMENT = None


def create_type(value: int, name: str) -> DataType:
    data = UInt8Data.from_int(value=value)
    fixed = DataType(data=data, value=value, name=name)
    DataType.cache(value=value, data_type=fixed)
    return fixed


DataType.COMMAND = create_type(value=0x00, name='Command')
DataType.COMMAND_RESPONSE = create_type(value=0x01, name='Command Response')
DataType.MESSAGE = create_type(value=0x02, name='Message')
DataType.MESSAGE_RESPONSE = create_type(value=0x03, name='Message Response')
DataType.MESSAGE_FRAGMENT = create_type(value=0x0A, name='Message Fragment')


class TransactionID(Data):

    ZERO = None  # TransactionID(data=Data(b'\0\0\0\0\0\0\0\0'))

    def __init__(self, data: Union[bytes, bytearray, ByteArray]):
        if isinstance(data, ByteArray):
            super().__init__(buffer=data.buffer, offset=data.offset, size=data.size)
        else:
            super().__init__(buffer=data)

    def __str__(self) -> str:
        clazz = self.__class__.__name__
        return '<%s: %s />' % (clazz, self.get_bytes())

    def __repr__(self) -> str:
        clazz = self.__class__.__name__
        return '<%s: %s />' % (clazz, self.get_bytes())

    #
    #   Factories
    #

    @classmethod
    def from_data(cls, data: Union[bytes, bytearray, ByteArray]):  # -> TransactionID
        if isinstance(data, cls):
            return data
        if isinstance(data, ByteArray):
            if data.size < 8:
                return None
            elif data.size > 8:
                data = data.slice(start=0, end=8)
        else:
            data_size = len(data)
            if data_size < 8:
                return None
            elif data_size > 8:
                data = data[:8]
        return cls(data=data)

    __number_lock = threading.Lock()
    __number_high = Convert.int32_from_data(data=Data.random(size=4))
    __number_low = Convert.int32_from_data(data=Data.random(size=4))

    @classmethod
    def generate(cls):  # -> TransactionID:
        with cls.__number_lock:
            if cls.__number_low < 0xFFFFFFFF:
                cls.__number_low += 1
            else:
                cls.__number_low = 0
                if cls.__number_high < 0xFFFFFFFF:
                    cls.__number_high += 1
                else:
                    cls.__number_high = 0
            hi = Convert.uint32data_from_value(value=cls.__number_high)
            lo = Convert.uint32data_from_value(value=cls.__number_low)
        return cls(data=hi.concat(lo))


TransactionID.ZERO = TransactionID(data=Data(b'\0\0\0\0\0\0\0\0'))



# -*- coding: utf-8 -*-
#
#   MTP: Message Transfer Protocol
#
#                                Written in 2020 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2020 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from typing import List, Optional

from ..ba import ByteArray, MutableData

from .protocol import DataType, TransactionID
from .package import Package


class Packer:

    def __init__(self, sn: TransactionID, pages: int):
        super().__init__()
        self.__sn = sn
        self.__pages = pages
        # assert sn is not None and pages > 1, 'pages error: %d' % pages
        self.__assembling: List[Package] = []      # message fragments
        self.__complete: Optional[Package] = None  # origin message package

    @property
    def sn(self) -> TransactionID:
        return self.__sn

    @property
    def completed(self) -> bool:
        return len(self.__assembling) == self.__pages

    @property
    def package(self) -> Optional[Package]:
        if self.__complete is None and self.completed:
            self.__complete = self.join(fragments=self.__assembling)
        return self.__complete

    @property
    def fragments(self) -> List[Package]:
        return self.__assembling

    def insert(self, fragment: Package) -> Optional[Package]:
        if self.__complete is not None:
            # already completed
            return self.__complete
        head = fragment.head
        assert head.sn == self.__sn, 'SN not match: %s, %s' % (head.sn, self.__sn)
        assert head.is_fragment, 'Package only for fragments: %s' % head.data_type
        assert head.pages == self.__pages, 'pages error: %d, %d' % (head.pages, self.__pages)
        assert head.index < self.__pages, 'index error: %d, %d' % (head.index, self.__pages)
        count = len(self.__assembling)
        index = count - 1
        while index >= 0:
            item = self.__assembling[index]
            if item.head.index < head.index:
                # got the position
                break
            elif item.head.index == head.index:
                # raise IndexError('duplicated: %s' % item.head)
                return self.__complete
            else:
                index -= 1
        # insert after the position
        self.__assembling.insert(index + 1, fragment)
        return self.package

    """
        Optimal Length for UDP Package Body
        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        MTU      : 576 bytes
        IP Head  : 20 bytes
        UDP Head : 8 bytes
        Header   : 12 bytes (excludes 'pages', 'index' and 'bodyLen')
        Reserved : 24 bytes (includes 'pages', 'index' and 'bodyLen')
    """
    OPTIMAL_BODY_LENGTH = 512

    @classmethod
    def join(cls, fragments: List[Package]) -> Optional[Package]:
        """
        Join sorted packages' body data together

        :param fragments: fragments sorted by index
        :return: original message package
        """
        count = len(fragments)
        assert count > 0, 'fragments count error: %d' % count
        first = fragments[0]
        if not first.is_fragment:
            # not fragments
            assert count == 1, 'packages error: %s, count=%d' % (first.head.data_type, count)
            return first
        sn = first.head.sn
        # get fragments count
        pages = first.head.pages
        assert pages == count, 'pages error: %d, %d' % (pages, count)
        # add message fragments part by part
        array: List[ByteArray] = []
        length = 0
        for index in range(count):
            item = fragments[index]
            assert item.head.is_fragment, 'data type should be fragment: %s' % item
            assert item.head.sn == sn, 'transaction ID not match: %s' % item
            assert item.head.pages == pages, 'pages error: %s' % item
            assert item.head.index == index, 'fragment missed: %d' % index
            array.append(item.body)
            length += item.body.size
        # join fragments
        body = MutableData(capacity=length)
        for index in range(count):
            body.append(source=array[index])
        if first.head.body_length < 0:
            # UDP (unlimited)
            assert first.head.body_length == -1, 'body length error: %d' % first.head.body_length
            return Package.new(data_type=DataType.MESSAGE, sn=sn, pages=1, index=0, body_length=-1, body=body)
        else:
            return Package.new(data_type=DataType.MESSAGE, sn=sn, pages=1, index=0, body_length=body.size, body=body)

    @classmethod
    def split(cls, package: Package) -> List[Package]:
        """
        Split large message package

        :param package: the large message package
        :return: message fragment packages
        """
        head = package.head
        body = package.body
        # check data type
        assert head.is_message, 'cannot split this type: %s' % head.data_type
        # split body
        fragments: List[ByteArray] = []
        pages = 1
        start = 0
        end = cls.OPTIMAL_BODY_LENGTH
        body_len = body.size
        while end < body_len:
            fragments.append(body.slice(start=start, end=end))
            pages += 1
            start = end
            end += cls.OPTIMAL_BODY_LENGTH
        if start > 0:
            fragments.append(body.slice(start=start))  # the tail
        else:
            fragments.append(body)  # the whole body (too small)
        # create packages with fragments
        sn = head.sn
        msg_fra = DataType.MESSAGE_FRAGMENT
        packages: List[Package] = []
        if pages == 1:
            # package too small, no need to split
            assert len(fragments) == 1, 'fragments error: %d' % len(fragments)
            packages.append(package)
        elif head.body_length < 0:
            # UDP (unlimited)
            assert head.body_length == -1, 'body length error: %d' % head.body_length
            for index in range(pages):
                data = fragments[index]
                pack = Package.new(data_type=msg_fra, sn=sn, pages=pages, index=index,
                                   body_length=-1, body=data)
                packages.append(pack)
        else:
            # TCP (should not happen)
            for index in range(pages):
                data = fragments[index]
                pack = Package.new(data_type=msg_fra, sn=sn, pages=pages, index=index,
                                   body_length=data.size, body=data)
                packages.append(pack)
        return packages

    @classmethod
    def sort(cls, fragments: List[Package]) -> List[Package]:
        """
        Sort the fragments with head.index

        :param fragments: fragments
        :return sorted fragments
        """
        fragments.sort(key=lambda pack: pack.head.index)
        return fragments



# -*- coding: utf-8 -*-
#
#   BA: Byte Array
#
#                                Written in 2020 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2020 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import ABC
from enum import IntEnum
from typing import Union


class ByteArray(ABC):

    @property
    def buffer(self) -> Union[bytes, bytearray]:
        """
        Get inner buffer

        :return: bytearray when it's mutable, or bytes for immutable
        """
        raise NotImplemented

    @property
    def offset(self) -> int:
        """
        Get view offset

        :return: 0 ~ (len(buffer) - 1)
        """
        raise NotImplemented

    @property
    def size(self) -> int:
        """
        Get view size

        :return: 0 ~ (len(buffer) - offset)
        """
        raise NotImplemented

    @property
    def hex_string(self) -> str:
        """
        Get Hex encoded string

        :return: hex string
        """
        raise NotImplemented

    def get_byte(self, index: int) -> int:
        """
        Get item value with position

        :param index: position
        :return: item value
        """
        raise NotImplemented

    def get_bytes(self, start: int = 0, end: int = None) -> bytes:
        """
        Get bytes from inner buffer with range [offset, offset + length)

        :param start: begin position relative to offset (include)
        :param end:   end position relative to offset (exclude)
        :return: data bytes
        """
        raise NotImplemented

    def slice(self, start: int, end: int = None):  # -> ByteArray:
        raise NotImplemented

    def concat(self, other):  # -> ByteArray:
        raise NotImplemented

    def find(self, sub, start: int = 0, end: int = None) -> int:
        """
        Search sub data/buffer within range [start, end)

        :param sub:   element value; or sub bytes; or sub array
        :param start: begin position relative to self.offset (include)
        :param end:   end position relative to self.offset (exclude)
        :return: first position where sub occurred; -1 on not found
        """
        raise NotImplemented


class MutableByteArray(ByteArray):

    def set_byte(self, index: int, value: int):
        """
        Change byte value at this position

        :param index: position
        :param value: byte value
        """
        raise NotImplemented

    def set_char(self, index: int, value: str):
        raise NotImplemented

    def update(self, index: int, source: Union[bytes, bytearray, ByteArray], start: int = 0, end: int = None):
        """
        Update values from source data/buffer with range [start, end)

        :param index:  update buffer from this relative position
        :param source: source data/buffer
        :param start:  source start position (include)
        :param end:    source end position (exclude)
        """
        raise NotImplemented

    def append(self, source: Union[bytes, bytearray, ByteArray], start: int = 0, end: int = None):
        """
        Append values from source data/buffer/value with range [start, end)

        :param source: source data/buffer
        :param start:  source start position (include)
        :param end:    source end position (exclude)
        """
        raise NotImplemented

    def insert(self, index: int, source: Union[bytes, bytearray, ByteArray], start: int, end: int = None):
        """
        Append values from source data/buffer/value with range [start, end)

        :param index:  insert buffer from this relative position
        :param source: source data/buffer
        :param start:  source start position (include)
        :param end:    source end position (exclude)
        """
        raise NotImplemented

    def remove(self, index: int) -> int:
        """
        Remove element at this position and return its value

        :param index: position
        :return: element value removed
        """
        raise NotImplemented

    def shift(self) -> int:
        """
        Remove element from the head position and return its value

        :return: element value at the first place
        """
        raise NotImplemented

    def pop(self) -> int:
        """
        Remove element from the tail position and return its value

        :return: element value at the last place
        """
        raise NotImplemented

    def push(self, element: int):
        """
        Append the element to the tail

        :param element: value
        """
        raise NotImplemented


"""
    Integer Data
    ~~~~~~~~~~~~
"""


class Endian(IntEnum):
    UNDEFINED = 0
    BIG_ENDIAN = 1  # Network Byte Order
    LITTLE_ENDIAN = 2  # Host Byte Order


class IntegerData(ByteArray):

    @property
    def endian(self) -> Endian:
        raise NotImplemented

    @property
    def value(self) -> int:
        raise NotImplemented

    @classmethod
    def get_value(cls, source: Union[bytes, bytearray, ByteArray],
                  start: int, size: int, endian: Endian) -> int:
        """
        Get integer value from source data/buffer within range [start, start+size)

        :param source: data or buffer
        :param start:  start position (include)
        :param size:   bytes count
        :param endian: Endian
        :return: int value
        """
        assert start >= 0 and size > 0, 'range error: start=%d, size=%d' % (start, size)
        if isinstance(source, ByteArray):
            start += source.offset
            source = source.buffer
        assert (start + size) <= len(source), 'length error: start=%d, size=%d, len=%d' % (start, size, len(source))
        result = 0
        if endian == Endian.LITTLE_ENDIAN:
            # [12 34 56 78] => 0x78563412
            pos = start + size - 1
            while pos >= start:
                result = (result << 8) | (source[pos] & 0xFF)
                pos -= 1
        elif endian == Endian.BIG_ENDIAN:
            # [12 34 56 78] => 0x12345678
            pos = start
            end = start + size
            while pos < end:
                result = (result << 8) | (source[pos] & 0xFF)
                pos += 1
        return result

    @classmethod
    def set_value(cls, value: int, data: Union[bytes, bytearray, MutableByteArray],
                  start: int, size: int, endian: Endian):
        """
        Set integer value into data/buffer
        """
        if isinstance(data, MutableByteArray):
            data.set_byte(index=(start + size - 1), value=0x00)  # for expanding spaces
            start += data.offset
            data = data.buffer
        if endian == Endian.LITTLE_ENDIAN:
            # 0x12345678 => [78 56 34 12]
            pos = start
            end = start + size
            while pos < end:
                data[pos] = value & 0xFF
                value >>= 8
                pos += 1
        elif endian == Endian.BIG_ENDIAN:
            # 0x12345678 => [12 34 56 78]
            pos = start + size - 1
            while pos >= start:
                data[pos] = value & 0xFF
                value >>= 8
                pos -= 1


#
#   Global Helpers
#
global_instances = {
    'data_helper': None,
    'mutable_helper': None,
    'integer_helper': None,
}


def get_data_helper():
    helper = global_instances['data_helper']
    from .helper import ByteArrayHelper
    assert isinstance(helper, ByteArrayHelper)
    return helper


def set_data_helper(helper):
    global_instances['data_helper'] = helper


def get_mutable_helper():
    helper = global_instances['mutable_helper']
    from .helper import MutableByteArrayHelper
    assert isinstance(helper, MutableByteArrayHelper)
    return helper


def set_mutable_helper(helper):
    global_instances['mutable_helper'] = helper


def get_integer_helper():
    helper = global_instances['integer_helper']
    from .helper import IntegerHelper
    assert isinstance(helper, IntegerHelper)
    return helper


def set_integer_helper(helper):
    global_instances['integer_helper'] = helper



# -*- coding: utf-8 -*-
#
#   BA: Byte Array
#
#                                Written in 2020 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2020 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from typing import Union, Optional

from .utils import varint_from_buffer, varint_to_buffer
from .array import get_integer_helper
from .array import ByteArray, IntegerData, Endian
from .data import Data


class IntData(Data, IntegerData):

    def __init__(self, data: Union[bytes, bytearray, ByteArray], value: int, endian: Endian = Endian.UNDEFINED):
        if isinstance(data, ByteArray):
            super().__init__(buffer=data.buffer, offset=data.offset, size=data.size)
        else:
            super().__init__(buffer=data)
        self.__value = value
        self.__endian = endian

    def __eq__(self, other) -> bool:
        if isinstance(other, IntegerData):
            return self.__value == other.value
        elif isinstance(other, int):
            return self.__value == other
        else:
            return super().__eq__(other=other)

    def __ne__(self, other) -> bool:
        if isinstance(other, IntegerData):
            return self.__value != other.value
        elif isinstance(other, int):
            return self.__value != other
        else:
            return super().__ne__(other=other)

    def __hash__(self) -> int:
        return hash(self.__value)

    def __str__(self) -> str:
        return str(self.__value)

    def __repr__(self) -> str:
        return str(self.__value)

    @property  # Override
    def endian(self) -> Endian:
        return self.__endian

    @property  # Override
    def value(self) -> int:
        return self.__value


class UInt8Data(IntData):
    """
        Unsigned Char (8-bytes)
    """
    ZERO = None  # UInt8Data.from_int(value=0)

    def __init__(self, data: Union[bytes, bytearray, ByteArray], value: int):
        super().__init__(data=data, value=value, endian=Endian.UNDEFINED)

    @classmethod
    def from_data(cls, data: Union[bytes, bytearray, ByteArray]) -> Optional[IntegerData]:
        if isinstance(data, cls):
            return data
        elif isinstance(data, ByteArray):
            size = data.size
            if size < 1:
                return None
            elif size > 1:
                data = data.slice(start=0, end=1)
            value = data.get_byte(index=0)
        else:
            assert isinstance(data, bytes) or isinstance(data, bytearray), 'uint8 data error: %s' % data
            size = len(data)
            if size < 1:
                return None
            elif size > 1:
                data = data[0:1]
            value = data[0]
        return cls(data=data, value=value)

    @classmethod
    def from_int(cls, value: int):
        data = bytearray(1)
        data[0] = value & 0xFF
        return cls(data=data, value=value)


class UInt16Data(IntData):
    """
        Unsigned Short Integer (16-bytes)
    """
    ZERO = None  # UInt16Data.from_int(value=0, endian=Endian.BIG_ENDIAN)

    @classmethod
    def from_data(cls, data: Union[bytes, bytearray, ByteArray], endian: Endian) -> Optional[IntegerData]:
        if isinstance(data, cls):
            return data
        elif isinstance(data, ByteArray):
            size = data.size
            if size < 2:
                return None
            elif size > 2:
                data = data.slice(start=0, end=2)
            value = get_integer_helper().get_value(buffer=data.buffer, offset=data.offset, size=2, endian=endian)
        else:
            assert isinstance(data, bytes) or isinstance(data, bytearray), 'uint16 data error: %s' % data
            size = len(data)
            if size < 2:
                return None
            elif size > 2:
                data = data[0:2]
            value = get_integer_helper().get_value(buffer=data, offset=0, size=2, endian=endian)
        return cls(data=data, value=value, endian=endian)

    @classmethod
    def from_int(cls, value: int, endian: Endian):
        buffer = bytearray(2)
        get_integer_helper().set_value(value=value, buffer=buffer, offset=0, size=2, endian=endian)
        return cls(data=buffer, value=value, endian=endian)


class UInt32Data(IntData):
    """
        Unsigned Integer (32-bytes)
    """
    ZERO = None  # UInt32Data.from_int(value=0, endian=Endian.BIG_ENDIAN)

    @classmethod
    def from_data(cls, data: Union[bytes, bytearray, ByteArray], endian: Endian) -> Optional[IntegerData]:
        if isinstance(data, cls):
            return data
        elif isinstance(data, ByteArray):
            size = data.size
            if size < 4:
                return None
            elif size > 4:
                data = data.slice(start=0, end=4)
            value = get_integer_helper().get_value(buffer=data.buffer, offset=data.offset, size=4, endian=endian)
        else:
            assert isinstance(data, bytes) or isinstance(data, bytearray), 'uint32 data error: %s' % data
            size = len(data)
            if size < 4:
                return None
            elif size > 4:
                data = data[0:4]
            value = get_integer_helper().get_value(buffer=data, offset=0, size=4, endian=endian)
        return cls(data=data, value=value, endian=endian)

    @classmethod
    def from_int(cls, value: int, endian: Endian):
        buffer = bytearray(4)
        get_integer_helper().set_value(value=value, buffer=buffer, offset=0, size=4, endian=endian)
        return cls(data=buffer, value=value, endian=endian)


class VarIntData(IntData):
    """
        Variable Integer
    """
    ZERO = None  # VarIntData.from_int(value=0)

    def __init__(self, data: Union[bytes, bytearray, ByteArray], value: int):
        super().__init__(data=data, value=value, endian=Endian.UNDEFINED)

    @classmethod
    def from_data(cls, data: Union[bytes, bytearray, ByteArray]) -> Optional[IntegerData]:
        if isinstance(data, cls):
            return data
        elif isinstance(data, ByteArray):
            value, size = varint_from_buffer(buffer=data.buffer, offset=data.offset, size=data.size)
            if size == 0:
                return None
            elif size < data.size:
                data = data.slice(start=0, end=size)
            return cls(data=data, value=value)
        else:
            assert isinstance(data, bytes) or isinstance(data, bytearray), 'varint data error: %s' % data
            value, size = varint_from_buffer(buffer=data, offset=0, size=len(data))
            if size == 0:
                return None
            data = Data(buffer=data, offset=0, size=size)
            return cls(data=data, value=value)

    @classmethod
    def from_int(cls, value: int):
        # maximum 8 bytes: 7 * 9 < 8 * 8 < 7 * 10
        buffer = bytearray(10)
        size = varint_to_buffer(value=value, buffer=buffer, offset=0, size=10)
        data = Data(buffer=buffer, offset=0, size=size)
        return cls(data=data, value=value)


#
#  Zero Values
#

UInt8Data.ZERO = UInt8Data.from_int(value=0)
UInt16Data.ZERO = UInt16Data.from_int(value=0, endian=Endian.BIG_ENDIAN)
UInt32Data.ZERO = UInt32Data.from_int(value=0, endian=Endian.BIG_ENDIAN)
VarIntData.ZERO = VarIntData.from_int(value=0)



# -*- coding: utf-8 -*-
#
#   BA: Byte Array
#
#                                Written in 2021 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2021 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import ABC, abstractmethod
from typing import Optional, Union

from .utils import array_concat, array_find
from .utils import array_set, array_update, array_insert, array_remove
from .utils import int_from_buffer, int_to_buffer
from .array import set_data_helper, set_mutable_helper, set_integer_helper
from .array import ByteArray, MutableByteArray, Endian
from .data import Data


class ByteArrayHelper(ABC):

    @abstractmethod
    def adjust(self, index: int, size: int) -> int:
        """ Adjust the index with range [0, size) """
        raise NotImplemented

    @abstractmethod
    def adjust_e(self, index: int, size: int) -> int:
        """ Adjust the index with range [0, size), throws when index too small """
        raise NotImplemented

    @abstractmethod
    def concat(self, left: ByteArray, right: ByteArray) -> ByteArray:
        """ left + right """
        raise NotImplemented

    @abstractmethod
    def find(self, sub: ByteArray, data: ByteArray, start: int, end: int) -> int:
        """ Search sub data """
        raise NotImplemented


class MutableByteArrayHelper(ABC):

    @abstractmethod
    def set(self, value: int, index: int, data: MutableByteArray) -> int:
        """ Set value to data at index, return new size """
        raise NotImplemented

    @abstractmethod
    def update(self, src: ByteArray, index: int, data: MutableByteArray) -> int:
        """ Update src to data at index, return new size """
        raise NotImplemented

    @abstractmethod
    def insert(self, src: ByteArray, index: int, data: MutableByteArray) -> int:
        """ Insert src to data at index, return new size """
        raise NotImplemented

    @abstractmethod
    def remove(self, index: int, data: MutableByteArray) -> (Optional[int], int, int):
        """ Remove element at index and return its value, new offset & size """
        raise NotImplemented


class IntegerHelper(ABC):

    @abstractmethod
    def get_value(self, buffer: Union[bytes, bytearray], offset: int, size: int, endian: Endian) -> int:
        """ Get integer value from data with range [offset, offset + size) """
        raise NotImplemented

    @abstractmethod
    def set_value(self, value: int, buffer: Union[bytearray], offset: int, size: int, endian: Endian):
        """ Set integer value into data with range [offset, offset + size) """
        raise NotImplemented


#
#   Implementations
#


class DefaultByteArrayHelper(ByteArrayHelper):
    """ Default ByteArrayHelper """

    # Override
    def adjust(self, index: int, size: int) -> int:
        if index < 0:
            index += size  # count from right hand
            if index < 0:
                return 0  # too small
        elif index > size:
            return size  # too big
        return index

    # Override
    def adjust_e(self, index: int, size: int) -> int:
        if index < 0:
            index += size  # count from right hand
            if index < 0:
                # too small
                raise IndexError('error index: %d, size: %d' % (index - size, size))
        return index

    # Override
    def concat(self, left: ByteArray, right: ByteArray) -> ByteArray:
        buffer, offset, size = array_concat(left_buffer=left.buffer, left_offset=left.offset, left_size=left.size,
                                            right_buffer=right.buffer, right_offset=right.offset, right_size=right.size)
        return Data(buffer=buffer, offset=offset, size=size)

    # Override
    def find(self, sub: ByteArray, data: ByteArray, start: int, end: int) -> int:
        if 0 < start or end < data.size:
            # slice
            data = Data(buffer=data.buffer, offset=(data.offset+start), size=(end-start))
        # searching within the range [start, end)
        pos = array_find(sub_buffer=sub.buffer, sub_offset=sub.offset, sub_size=sub.size,
                         buffer=data.buffer, offset=data.offset, size=data.size)
        if pos == -1:
            return -1
        else:
            return pos + start


class DefaultMutableByteArrayHelper(MutableByteArrayHelper):
    """ Default MutableByteArrayHelper """

    # Override
    def set(self, value: int, index: int, data: MutableByteArray) -> int:
        return array_set(index=index, value=value,
                         buffer=data.buffer, offset=data.offset, size=data.size)

    # Override
    def update(self, src: ByteArray, index: int, data: MutableByteArray) -> int:
        return array_update(index=index, src=src.buffer, src_offset=src.offset, src_size=src.size,
                            buffer=data.buffer, offset=data.offset, size=data.size)

    # Override
    def insert(self, src: ByteArray, index: int, data: MutableByteArray) -> int:
        return array_insert(index=index, src=src.buffer, src_offset=src.offset, src_size=src.size,
                            buffer=data.buffer, offset=data.offset, size=data.size)

    # Override
    def remove(self, index: int, data: MutableByteArray) -> (Optional[int], int, int):
        return array_remove(index=index, buffer=data.buffer, offset=data.offset, size=data.size)


class DefaultIntegerHelper(IntegerHelper):
    """ Default IntegerHelper """

    # Override
    def get_value(self, buffer: Union[bytes, bytearray], offset: int, size: int, endian: Endian) -> int:
        return int_from_buffer(buffer=buffer, offset=offset, size=size, endian=endian)

    # Override
    def set_value(self, value: int, buffer: Union[bytearray], offset: int, size: int, endian: Endian):
        int_to_buffer(value=value, buffer=buffer, offset=offset, size=size, endian=endian)


# set default helpers
set_data_helper(helper=DefaultByteArrayHelper())
set_mutable_helper(helper=DefaultMutableByteArrayHelper())
set_integer_helper(helper=DefaultIntegerHelper())



# -*- coding: utf-8 -*-
#
#   TLV: Tag Length Value
#
#                                Written in 2020 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2020 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from typing import Union, Optional

from .array import get_data_helper, get_mutable_helper
from .array import ByteArray, MutableByteArray
from .data import adjust_positions
from .data import Data


class MutableData(Data, MutableByteArray):

    def __init__(self, buffer: bytearray = None, offset: int = 0, size: int = None, capacity: int = 4):
        if buffer is None:
            assert capacity > 0, 'invalid capacity: %d' % capacity
            buffer = bytearray(capacity)  # create empty buffer with capacity
            offset = 0
            size = 0
        super().__init__(buffer=buffer, offset=offset, size=size)

    # Override
    def set_byte(self, index: int, value: int):
        index = get_data_helper().adjust_e(index=index, size=self._size)
        size = get_mutable_helper().set(value=value, index=index, data=self)
        if size > self._size:
            self._size = size

    # Override
    def set_char(self, index: int, value: str):
        assert len(value) == 1, 'char value error: %s' % value
        self.set_byte(index=index, value=ord(value[0]))

    # Override
    def update(self, index: int, source: Union[bytes, bytearray, ByteArray], start: int = 0, end: int = None):
        index = get_data_helper().adjust_e(index=index, size=self._size)
        source = get_sub_data(data=source, start=start, end=end)
        size = get_mutable_helper().update(src=source, index=index, data=self)
        if size > self._size:
            self._size = size

    # Override
    def append(self, source: Union[bytes, bytearray, ByteArray], start: int = 0, end: int = None):
        source = get_sub_data(data=source, start=start, end=end)
        size = get_mutable_helper().insert(src=source, index=self._size, data=self)
        if size > self._size:
            self._size = size

    # Override
    def insert(self, index: int, source: Union[bytes, bytearray, ByteArray], start: int, end: int = None):
        index = get_data_helper().adjust_e(index=index, size=self._size)
        source = get_sub_data(data=source, start=start, end=end)
        size = get_mutable_helper().insert(src=source, index=index, data=self)
        if size > self._size:
            self._size = size

    # Override
    def remove(self, index: int) -> int:
        index = get_data_helper().adjust_e(index=index, size=self._size)
        value, offset, size = get_mutable_helper().remove(index=index, data=self)
        if offset > self._offset:
            self._offset = offset
        if size < self._size:
            self._size = size
        return value

    # Override
    def shift(self) -> int:
        if self._size < 1:
            raise IndexError('data empty!')
        erased = self._buffer[self._offset]
        self._offset += 1
        self._size -= 1
        return erased

    # Override
    def pop(self) -> int:
        if self._size < 1:
            raise IndexError('data empty!')
        self._size -= 1
        return self._buffer[self._offset + self._size]

    # Override
    def push(self, element: int):
        self.set_byte(index=self._size, value=element)


def get_sub_data(data: Union[bytes, bytearray, ByteArray], start: int, end: Optional[int]) -> ByteArray:
    if isinstance(data, ByteArray):
        return data.slice(start=start, end=end)
    else:
        start, end = adjust_positions(size=len(data), start=start, end=end)
        return Data(buffer=data, offset=start, size=(end - start))



# -*- coding: utf-8 -*-
#
#   MTP: Message Transfer Protocol
#
#                                Written in 2020 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2020 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from typing import Union

from ..ba import ByteArray, Data

from .protocol import DataType, TransactionID
from .header import Header


class Package(Data):

    def __init__(self, data: Union[bytes, bytearray, ByteArray], head: Header, body: ByteArray):
        if isinstance(data, ByteArray):
            super().__init__(buffer=data.buffer, offset=data.offset, size=data.size)
        else:
            super().__init__(buffer=data)
        self.__head = head
        self.__body = body

    def __str__(self) -> str:
        clazz = self.__class__.__name__
        return '<%s: head=%s body_len=%d />' % (clazz, self.head, self.body.size)

    def __repr__(self) -> str:
        clazz = self.__class__.__name__
        return '<%s: head=%s body_len=%d />' % (clazz, self.head, self.body.size)

    @property
    def head(self) -> Header:
        return self.__head

    @property
    def body(self) -> ByteArray:
        return self.__body

    @property
    def is_response(self) -> bool:
        return self.__head.data_type.is_response

    @property
    def is_fragment(self) -> bool:
        return self.__head.data_type.is_fragment

    @property
    def is_command(self) -> bool:
        return self.__head.data_type.is_command

    @property
    def is_command_response(self) -> bool:
        return self.__head.data_type.is_command_response

    @property
    def is_message(self) -> bool:
        return self.__head.data_type.is_message

    @property
    def is_message_response(self) -> bool:
        return self.__head.data_type.is_message_response

    @property
    def is_message_fragment(self) -> bool:
        return self.__head.data_type.is_message_fragment

    @classmethod
    def parse(cls, data: ByteArray):  # -> Package
        # get package head
        head = Header.parse(data=data)
        if head is None:
            # raise AssertionError('package head error')
            return None
        # check lengths
        head_len = head.size
        body_len = head.body_length
        if body_len < 0:
            # UDP (unlimited)
            assert body_len == -1, 'body length error: %d' % body_len
        else:
            # TCP
            pack_len = head_len + body_len
            data_len = data.size
            if data_len < pack_len:
                # raise IndexError('package length error: %s' % data)
                return None
            elif data_len > pack_len:
                data = data.slice(start=0, end=pack_len)
        # get package body
        body = data.slice(start=head_len)
        return cls(data=data, head=head, body=body)

    @classmethod
    def new(cls, data_type: DataType, sn: TransactionID = None, pages: int = 1, index: int = 0, body_length: int = -1,
            body: ByteArray = None):
        # create package with header
        head = Header.new(data_type=data_type, sn=sn, pages=pages, index=index, body_length=body_length)
        if body is None:
            data = head
            body = Data.ZERO
        else:
            data = head.concat(body)
        return cls(data=data, head=head, body=body)



# -*- coding: utf-8 -*-
#
#   BA: Byte Array
#
#                                Written in 2020 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2020 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from typing import Union, Optional

from .utils import hex_encode, random_bytes
from .utils import array_equal, array_hash
from .array import get_data_helper
from .array import ByteArray, MutableByteArray


class Data(ByteArray):
    """
        Data View
    """
    ZERO = None  # Data(buffer=b'')

    def __init__(self, buffer: Union[bytes, bytearray], offset: int = 0, size: int = None):
        """
        Create data view with range [start, end)

        :param buffer: bytes, bytearray or another data view
        :param offset: data view offset
        :param size:   data view size
        """
        super().__init__()
        self._buffer = buffer
        self._offset = offset
        if size is None:
            self._size = len(buffer) - offset
        else:
            self._size = size

    @property  # Override
    def buffer(self) -> Union[bytes, bytearray]:
        return self._buffer

    @property  # Override
    def offset(self) -> int:
        return self._offset

    @property  # Override
    def size(self) -> int:
        return self._size

    def __eq__(self, other) -> bool:
        if other is None:
            return self._size == 0
        elif other is self:
            return True
        elif isinstance(other, ByteArray):
            return array_equal(left_buffer=self._buffer, left_offset=self._offset, left_size=self._size,
                               right_buffer=other.buffer, right_offset=other.offset, right_size=other.size)
        elif isinstance(other, bytes) or isinstance(other, bytearray):
            return array_equal(left_buffer=self._buffer, left_offset=self._offset, left_size=self._size,
                               right_buffer=other, right_offset=0, right_size=len(other))
        else:
            return False

    def __ne__(self, other) -> bool:
        if other is None:
            return self._size > 0
        elif other is self:
            return False
        elif isinstance(other, ByteArray):
            return not array_equal(left_buffer=self._buffer, left_offset=self._offset, left_size=self._size,
                                   right_buffer=other.buffer, right_offset=other.offset, right_size=other.size)
        elif isinstance(other, bytes) or isinstance(other, bytearray):
            return not array_equal(left_buffer=self._buffer, left_offset=self._offset, left_size=self._size,
                                   right_buffer=other, right_offset=0, right_size=len(other))
        else:
            return True

    def __len__(self) -> int:
        return self._size

    def __hash__(self) -> int:
        return array_hash(buffer=self._buffer, offset=self._offset, size=self._size)

    def __str__(self) -> str:
        data = self.get_bytes()
        # return data.decode('utf-8')
        return str(data)

    def __repr__(self) -> str:
        data = self.get_bytes()
        # return data.decode('utf-8')
        return str(data)

    def __add__(self, other):
        """ self + other """
        return concat(left=self, right=other)

    @property  # Override
    def hex_string(self) -> str:
        data = self.get_bytes()
        return hex_encode(data=data)

    # Override
    def get_byte(self, index: int) -> int:
        index = get_data_helper().adjust_e(index=index, size=self._size)
        if index >= self._size:
            # too big
            raise IndexError('error index: %d, size: %d' % (index, self._size))
        return self._buffer[self._offset + index]

    # Override
    def get_bytes(self, start: int = 0, end: int = None) -> bytes:
        start, end = adjust_positions(size=self._size, start=start, end=end)
        sub = get_slice(data=self, start=start, end=end)
        if isinstance(sub, bytearray):
            return bytes(sub)
        else:
            return sub

    # Override
    def slice(self, start: int, end: int = None) -> ByteArray:
        start, end = adjust_positions(size=self._size, start=start, end=end)
        if start == 0 and end == self._size:
            # whole data
            return self
        elif start < end:
            # sub view
            return Data(buffer=self._buffer, offset=(self._offset+start), size=(end-start))
        else:
            # error
            return self.ZERO

    # Override
    def concat(self, other: Union[bytes, bytearray, ByteArray]) -> ByteArray:
        return concat(left=self, right=other)

    # Override
    def find(self, sub: Union[bytes, bytearray, ByteArray], start: int = 0, end: int = None) -> int:
        if not isinstance(sub, ByteArray):
            sub = Data(buffer=sub)
        start, end = adjust_positions(size=self._size, start=start, end=end)
        return get_data_helper().find(sub=sub, data=self, start=start, end=end)

    def mutable_copy(self) -> MutableByteArray:
        data = get_slice(data=self, start=0, end=self._size)
        from .mutable import MutableData
        return MutableData(buffer=data)

    @classmethod
    def random(cls, size: int) -> ByteArray:
        return cls(buffer=random_bytes(size=size))


def convert(data: Union[bytes, bytearray, ByteArray]) -> ByteArray:
    if data is None:
        return Data.ZERO
    elif isinstance(data, ByteArray):
        return data
    else:
        return Data(buffer=data)


def concat(left: Union[bytes, bytearray, ByteArray], right: Union[bytes, bytearray, ByteArray]) -> ByteArray:
    left = convert(data=left)
    right = convert(data=right)
    if left.size == 0:
        return right
    elif right.size == 0:
        return left
    else:
        return get_data_helper().concat(left=left, right=right)


# get slice with range [start, end)
def get_slice(data: ByteArray, start: int, end: int) -> Union[bytes, bytearray]:
    if start >= end:
        return b''
    start += data.offset
    end += data.offset
    if start < 0 or end > (data.offset + data.size):
        raise IndexError('range error: [%d, %d), offset=%d, size=%d' % (start, end, data.offset, data.size))
    return data.buffer[start:end]


def adjust_positions(size: int, start: int, end: Optional[int]) -> (int, int):
    helper = get_data_helper()
    start = helper.adjust(index=start, size=size)
    if end is None:
        end = size
    else:
        end = helper.adjust(index=end, size=size)
    return start, end


# empty data
Data.ZERO = Data(buffer=b'')
